import os

from pymongo import MongoClient

connection_string = os.environ.get("MONGO_URL", "mongodb://localhost:27017/oatk")
server, db_name   = connection_string.rsplit("/", 1)
client            = MongoClient(server)
db                = client[db_name]






# https://github.com/lepture/authlib/blob/master/authlib/integrations/sqla_oauth2/client_mixin.py

# client = {
#   "id"
#   "user_id"
#   "user" : {
#     "id"
#     "username"
#     get_user_id
#     str
#   }
#
#   "client_id"
#   "client_secret"
#   "client_id_issued_at"
#   "client_secret_expires_at"
#   "client_metadata"
# }

# https://github.com/lepture/authlib/blob/master/authlib/integrations/sqla_oauth2/tokens_mixins.py

# code = {
#   "id"
#   "user_id"
#   "user" : {
#     "id"
#     "username"
#     get_user_id
#     str
#   }
#
#     "code"
#     "client_id"
#     "redirect_uri"
#     "response_type"
#     "scope"
#     "nonce"
#     "auth_time"
#     "code_challenge"
#     "code_challenge_method"
# }

# token = {
#   "id"
#   "user_id"
#   "user" : {
#     "id"
#     "username"
#     get_user_id
#     str
#   }
#
#     "client_id"
#     "token_type"
#     "access_token"
#     "refresh_token"
#     "scope"
#     "issued_at"
#     "access_token_revoked_at"
#     "refresh_token_revoked_at"
#     "expires_in"
# }



from flask import Flask, request, Response
from flask_restful import Resource, Api
from flask_cors import CORS

from oatk import OAuthToolkit

app = Flask(__name__)
CORS(app, resources={r"*": {"origins": "*"}})
api = Api(app)

oatk = OAuthToolkit()
oatk.with_private("private_key.pem")
oatk.with_jwks("certs.json")

@app.route("/")
@oatk.authenticated
def hello_world():
  return "<p>Hello, World!</p>"

class HelloWorld(Resource):
  @oatk.authenticated_with_claims(username="xtof")
  def get(self):
    return {"hello": "world"}

api.add_resource(HelloWorld, "/api/hello")

class Token(Resource):
  def get(self):
    claims = {}
    for claim, value in request.args.items():
      if "," in value:
        value = value.split(",")
      claims[claim] = value
    token = oatk.claims(**claims).token
    return Response(response=token, status=200, mimetype="plain/text")

api.add_resource(Token, "/api/token")



import logging
logger = logging.getLogger(__name__)

import os
import sys

# load the environment variables for this setup
from dotenv import load_dotenv, find_dotenv
load_dotenv(find_dotenv(usecwd=True))

LOG_LEVEL = os.environ.get("LOG_LEVEL") or "INFO"

# setup logging infrastructure

logging.getLogger("urllib3").setLevel(logging.WARN)

FORMAT  = "[%(asctime)s] [%(name)s] [%(levelname)s] %(message)s"
DATEFMT = "%Y-%m-%d %H:%M:%S %z"

logging.basicConfig(level=LOG_LEVEL, format=FORMAT, datefmt=DATEFMT)
formatter = logging.Formatter(FORMAT, DATEFMT)
logging.getLogger().handlers[0].setFormatter(formatter)

from fire import Fire

from oatk import OAuthToolkit

def cli():
  Fire(OAuthToolkit, name="oatk")
  
if __name__ == '__main__':
  cli()



from oatk import OAuthToolkit

# create token

oauth = OAuthToolkit()
oauth.with_private("private_key.pem")
oauth.with_public("public_key.pem")

oauth.claims(hello="world")

token = oauth.token
oauth.validate(token)

jwks = oauth.jwks

# validate use case

oauth2 = OAuthToolkit()
oauth2.with_jwks(jwks)
oauth2.validate(token)

# or as a chained one-liner
OAuthToolkit().with_jwks(jwks).validate(token)




import logging
logger = logging.getLogger(__name__)

import time
import random
import string
import uuid
import json

from flask import request, session, redirect
from flask import render_template, redirect, jsonify
from werkzeug.security import gen_salt

from . import server
from .db import db

@server.route("/", methods=["GET", "POST"])
def home():
  """
  The root/home page allows a user to login (POST) and manage its clients.
  Nothing OAuth to see here, just to server's own functionality.
  From this page a link to the /oauth/create-cient route is provided.
  """
  if request.method == "POST":
    username = request.form.get("username")
    user = db["users"].find_one({"username" : username})
    if not user:
      user = { "username": username, "_id" : str(uuid.uuid4()) }
      db["users"].insert_one(user)
    session["id"] = user["_id"]
    return redirect("/")

  # GET
  user = current_user()
  if user:
    clients = list(db["clients"].find({"user_id" : user["_id"]}))
    for client in clients:
      client["_id"] = str(client["_id"]) # turn Mongo ObjectID into string
    return render_template("home.html", user=user, clients=clients)
  else:
    return render_template("login.html")

@server.route("/oauth/create-client", methods=["GET", "POST"])
def create_client():
  """
  The Create Client page allows a user to create a new client registration.
  Client registration is required for an external application to interact and
  act on behalf of our user.
  In itself also no real OAuth shizzle here. Just administrative functionality.
  """
  user = current_user()
  if not user:
    return redirect("/")

  if request.method == "GET":
    return render_template(
      "create_client.html", client_id=gen_salt(24), client_secret=gen_salt(48)
    )

  # POST
  form = request.form
  try:
    db["clients"].insert_one({
      "client_id"           : form["client_id"],
      "user_id"             : user["_id"],
      "client_id_issued_at" : int(time.time()),
      "client_secret"       : form["client_secret"],
      "metadata" : {
        "client_name"               : form["client_name"],
        "client_uri"                : form["client_uri"],
        "grant_types"               : split_by_crlf(form["grant_type"]),
        "redirect_uris"             : split_by_crlf(form["redirect_uri"]),
        "response_types"            : split_by_crlf(form["response_type"]),
        "scope"                     : form["scope"],
        "allowed-origins"           : split_by_crlf(form["allowed_origins"]),
        "permission_groups"         : split_by_crlf(form["permission_groups"])
      }
    })
  except Exception as ex:
    logger.exception("failed to register client")
  return redirect("/")

@server.route("/oauth/authorize", methods=["GET", "POST"])
def authorize():
  """
  The first real OAuth route, and the first to hit in the flow. The application
  redirects the browser here to receive back an authoriation code. If the user
  isn't logged in yet, she will have to do so first. When logged in, she can
  effectively authorize the application to handle on its behalf.
  """
  if request.method == "POST":
    logger.info("processing post to auth route")
    username = request.form.get("username")
    logger.info(f"got {username}")
    if username:
      user = db["users"].find_one({"username" : username})
      logger.info(f"found user in db: {user}")
      if user:
        session["id"] = user["_id"]

  client_id = request.args["client_id"]
  scope     = request.args["scope"]

  # retrieve client for user
  user = current_user()
  logger.info(f"current_user: {user}")
  if user:
    client = db["clients"].find_one({
      "user_id"   : user["_id"],
      "client_id" : client_id
    })
    assert scope == "openid profile"
    assert request.args["response_type"] == "code"
    nonce = request.args["nonce"]
    logger.info(f"got client: {client}")
  else:
    return render_template("login.html", args=str(request.query_string.decode()))

  if "confirm" in request.form:
    logger.info("got confirmation: f{request.form['confirm']}")
    if request.form["confirm"]:
      logger.info("get positive confirmation... redirecting with code...")
      redirect_uri = client["metadata"]["redirect_uris"][0]
      code = generate_token()
      db["codes"].insert_one({
        "code"         : code,
        "client_id"    : client["client_id"],
        "user_id"      : client["user_id"],
        "redirect_uri" : redirect_uri,
        "scope"        : "openid"
      })
      goto = f"{redirect_uri}?code={code}"
      logger.debug(goto)
      return redirect(goto, 302)

  try:
    grants = client["metadata"]["grant_types"]
    assert "authorization_code" in grants
    grant = {
      "client"  : { "client_name" : client["metadata"]["client_name"] },
      "request" : { "scope"       : scope }
    }
  except Exception as error:
    logger.exception(error)
    return jsonify({"error" : str(error)})
  return render_template("authorize.html", grant=grant)

@server.route("/oauth/token", methods=["POST"])
def issue_token():
  try:
    code = db["codes"].find_one({"code" : request.json["code"] })
    client = db["clients"].find_one({
      "user_id"   : code["user_id"],
      "client_id" : code["client_id"]
    })
    user = db["users"].find_one({"_id" : code["user_id"]})
    username = user["username"]

    now = round(time.time())
    token = {
      "iat"              : now,
      "exp"              : now + 300,
      "auth_time"        : now,
      "jti"              : str(uuid.uuid4()),
      "iss"              : request.url_root.rstrip("/"),
      "sub"              : f"f:{str(uuid.uuid4())}:{client['metadata']['client_name']}",
      "typ"              : "Bearer",
      "azp"              : "development",
      "nonce"            : "nonce-TODO",
      "session_state"    : "52ad15e0-4afb-4b04-8897-8e08b262a73d",
      "acr"              : "0",
      "allowed-origins"  : client["metadata"]["allowed-origins"],
      "scope"            : "openid profile permission_groups",
      "sid"              : "52ad15e0-4afb-4b04-8897-8e08b262a73d",
      "permission_groups": client["metadata"]["permission_groups"],
      "username"         : username
    }
    
    encoded = server.oatk.claims(token).token
    return {
      "access_token": encoded,
      "expires_in": 300,
      "refresh_expires_in": 0,
      "token_type": "Bearer",
      "id_token": "TODO:id_token",
      "not-before-policy": str(time.time()),
      "scope": "openid permission_groups profile"
    }

  except Exception as ex:
    logger.exception("failed to provide token")
  return {}


@server.route("/oauth/userinfo")
@server.oatk.authenticated_with_claims(scope="profile")
def api_me():
  logger.warn("TODO implement actual userinfo")
  return { "hello" : "world", "STILL": "TODO" }

@server.route("/oauth/certs")
def certs():
  return json.dumps(server.oatk.jwks, indent=2)

@server.route("/oauth/well-known")
def well_known():
  me = request.url_root.rstrip("/")
  return json.dumps({
    "issuer"                 : f"{me}",
    "jwks_uri"               : f"{me}/oauth/certs",
    "authorization_endpoint" : f"{me}/oauth/authorize",
    "token_endpoint"         : f"{me}/oauth/token",
    "userinfo_endpoint"      : f"{me}/oauth/userinfo",
    "end_session_endpoint"   : f"{me}/oauth/logout",
    "registration_endpoint"  : f"{me}/oauth/create-client",
  }, indent=2)

@server.route("/oauth/logout")
def logout():
  session.pop("id", None)
  goto = "/"
  return redirect(goto, 302)

# helper functions

def current_user():
  if "id" in session:
    uid = session["id"]
    return db["users"].find_one({"_id" : uid})
  return None

def split_by_crlf(s):
  return [v for v in s.splitlines() if v]

UNICODE_ASCII_CHARACTER_SET = string.ascii_letters + string.digits

def generate_token(length=30, chars=UNICODE_ASCII_CHARACTER_SET):
  rand = random.SystemRandom()
  return ''.join(rand.choice(chars) for _ in range(length))



__version__ = "0.1.0"

import logging
logger = logging.getLogger(__name__)

import json
import uuid

import jwt
from authlib.jose import jwk

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

from functools import wraps
from flask import request, Response

import requests

from oatk import fake

try:
  from AppKit import NSPasteboard, NSStringPboardType
  pb = NSPasteboard.generalPasteboard()
except ModuleNotFoundError:
  logger.debug("No AppKit installed, so no MacOS clipboard support!")
  pb = None

class OAuthToolkit():
  def __init__(self):
    self._encoded     = None
    self._certs       = {}
    self._private_key = None
    self._public_key  = None
    self._alg         = "RS256"
    self._kid         = str(uuid.uuid4())
    self._claims      = {}
    self._client_id   = None
    
    self.server       = fake.server
    self.server.oatk  = self

  @property
  def version(self):
    return __version__

  def with_private(self, path):
    with open(path, "rb") as fp:
      self._private_key = serialization.load_pem_private_key(
        fp.read(),
        password=None,
        backend=default_backend()
      )
    return self

  def with_public(self, path):
    with open(path, "rb") as fp:
      self._public_key = serialization.load_pem_public_key(
        fp.read(),
        backend=default_backend()
      )
    self._certs = { self._kid : self._public_key }
    return self

  def using_provider(self, provider_url):
    try:
      config = json.loads(requests.get(provider_url).content)
    except:
      logger.exception("could not retrieve openid configuration")
      return
    try:
      self.with_jwks(requests.get(config["jwks_uri"]).content)
    except:
      logger.exception("could not import jwks")
      return
    logger.info(f"succesfully configured from {provider_url}")
    return self

  def with_client_id(self, client_id):
    self._client_id = client_id
    return self

  @property
  def jwks(self):
    return json.dumps({
      "keys" : [
        jwk.dumps(self._public_key, kty="RSA", alg=self._alg, kid=self._kid)
      ]
    }, indent=2)

  def with_jwks(self, path_or_string_or_obj):
    try:
      with open(path_or_string_or_obj) as fp:
        jwks = json.load(fp)
    except:
      try:
        jwks = json.loads(path_or_string_or_obj)
      except:
        jwks = path_or_string_or_obj
    assert isinstance(jwks, dict)
    self._certs = {
      key["kid"] : jwt.algorithms.RSAAlgorithm.from_jwk(json.dumps(key))
      for key in jwks["keys"]
    }
    if jwks["keys"]:
      self._kid = jwks["keys"][0]["kid"]
    return self
  
  def from_clipboard(self):
    encoded = pb.stringForType_(NSStringPboardType)
    if encoded[:6] == "Bearer":
      encoded = clip[7:]
    self._encoded = encoded.strip() # strip to remove trailing newline
    return self

  def from_file(self, path):
    with open(path) as fp:
      self._encoded = fp.read().strip() # strip to remove trailing newline
    return self

  def header(self, token=None):
    if not token:
      token = self._encoded
    return jwt.get_unverified_header(token)

  def claims(self, claimsdict=None, **claimset):
    if claimsdict is None: claimsdict = {}
    self._claims = claimset
    self._claims.update(claimsdict)
    return self

  @property
  def token(self):
    if self._private_key:
      return jwt.encode(
        self._claims, self._private_key, algorithm=self._alg,
        headers={ "kid": self._kid, "alg" : self._alg }
      )
    return None

  def validate(self, token=None):
    kid = self.header(token)["kid"]
    alg = self.header(token)["alg"]
    if not token:
      token = self._encoded
    jwt.decode( token, self._certs[kid], algorithms=[alg], audience=self._client_id )

  def decode(self, token=None):
    if not token:
      token = self._encoded
    return jwt.decode( token, options={"verify_signature": False} )

  def execute_authenticated(self, f, required_claims=None, *args, **kwargs):
    code = 403
    msg  = ""
    try:
      token = request.headers["Authorization"][7:]
      self.validate(token)
      if required_claims:
        claims = self.decode(token)
        for claim, value in required_claims.items():
          if not claim in claims:
            raise ValueError(f"required claim {claim} is missing")
          if callable(value):
            if not value(claims[claim]):
              raise ValueError(f"claim {claim} doesn't match required criteria")
          elif type(value) == list:
            if not value in claims[claim]:
              raise ValueError(f"claim {claim} is missing required value")
          elif value != claims[claim]:
            raise ValueError(f"claim {claim} doesn't equal required value")
      # authenticated -> execute
      return f(*args, **kwargs)
    except KeyError as ex:
      msg = "Missing Authorization"
      code = 401
    except ValueError as e:
      msg = str(e)
      logger.warning(msg)
    except Exception as e:
      msg = repr(e)
      logger.warning(f"unexpected exception: {msg}")
    return Response(msg, code)

  def authenticated(self, f):
    @wraps(f)
    def wrapper(*args, **kwargs):
      return self.execute_authenticated(f, *args, **kwargs)
    return wrapper

  def authenticated_with_claims(self, **required_claims):
    def decorator(f):
      @wraps(f)
      def wrapper(*args, **kwargs):
        return self.execute_authenticated(f, required_claims, *args, **kwargs)
      return wrapper
    return decorator



# load the environment variables for this setup from .env file
from dotenv import load_dotenv, find_dotenv
from pathlib import Path
load_dotenv(find_dotenv())
load_dotenv(find_dotenv(".env.local"))

import logging
logger = logging.getLogger(__name__)

import os

LOG_LEVEL = os.environ.get("LOG_LEVEL") or "INFO"

# setup logging infrastructure

logging.getLogger("urllib3").setLevel(logging.WARN)

FORMAT  = "[%(asctime)s] [%(name)s] [%(levelname)s] %(message)s"
DATEFMT = "%Y-%m-%d %H:%M:%S %z"

logging.basicConfig(level=LOG_LEVEL, format=FORMAT, datefmt=DATEFMT)
formatter = logging.Formatter(FORMAT, DATEFMT)
logging.getLogger().handlers[0].setFormatter(formatter)

from flask import Flask, render_template, Response
from flask_restful import Resource, Api

import oatk.js
from oatk import OAuthToolkit

server = Flask(__name__)

# route to load web app
@server.route("/", methods=["GET"])
def home():
  return render_template(
    "home.html",
    OAUTH_PROVIDER=os.environ["OAUTH_PROVIDER"],
    OAUTH_CLIENT_ID=os.environ["OAUTH_CLIENT_ID"]
  )

# route for oatk.js from the oatk package
@server.route("/oatk.js", methods=["GET"])
def oatk_script():
  return Response(oatk.js.as_src(), mimetype="application/javascript")

# API set up
api = Api(server)

# setup oatk
auth = OAuthToolkit()
auth.using_provider(os.environ["OAUTH_PROVIDER"]);
auth.with_client_id(os.environ["OAUTH_CLIENT_ID"])

def validate_name(name):
  return name == "Christophe VG"

class HelloWorld(Resource):
  @auth.authenticated_with_claims(name=validate_name)
  def get(self):
    return {"hello": "world"}

api.add_resource(HelloWorld, "/api/hello")
